/*
  The encoder can be used to create a h264 stream out of a sequence of images.
  In order to be able to use the encoder, it is neccessary to call the function setup_encoder(...) during the setup of the vulkan device.
  Besides that, the requested vulkan version has to be set to Vulkan 1.1 during the setup of the vulkan instance.
  A frame can be submitted for encoding using the function encode_frame(...).
  After the completion of an encode task, a callback function is executed to which the resulting ouput of the task is passed.
  Example:

   //During on_setup_instance(...)
   req_api_version = lava::api_version::v1_1;

   //During on_setup_device(...)
   setup_encoder(...);

   //During creation and submission of a frame
   Encoder::Ptr encoder = make_encoder();

   encoder->create(...);

   encoder->encoder_frame(..., image, ..., [](const std::span<uint8_t>& content, EncoderFrameType type)
   {
        //Use content
   });
*/

#pragma once
#include <liblava/lava.hpp>
#include <asio.hpp>
#include <optional>
#include <vector>
#include <span>
#include <memory>
#include <mutex>
#include <functional>
#include <optional>

#include "extern_fence.hpp"

enum EncoderMode
{
    ENCODER_MODE_CONSTANT_BITRATE,
    ENCODER_MODE_CONSTANT_QUALITY
};

enum EncoderFrameType
{
    ENCODER_FRAME_TYPE_CONTROL,
    ENCODER_FRAME_TYPE_CONFIG,
    ENCODER_FRAME_TYPE_FRAME
};

typedef std::function<void(const std::span<uint8_t>& content, EncoderFrameType type)> EncoderCallback;

struct EncoderFrame
{
public:
    typedef std::shared_ptr<EncoderFrame> Ptr;

public:
    VmaAllocation input_memory_combined;
    VmaAllocation input_memory_luma;
    VmaAllocation input_memory_chroma;
    VmaAllocation input_memory_chroma_subsampled;

    VkImage input_luma = nullptr;
    VkImage input_chroma = nullptr;
    VkImage input_chroma_subsampled = nullptr;
    VkImage input_combined = nullptr;

    VkImageView input_view_luma = nullptr;
    VkImageView input_view_chroma = nullptr;
    VkImageView input_view_chroma_subsampled = nullptr;
    VkImageView input_view_combined = nullptr;

    VmaAllocation slot_memory_luma;
    VmaAllocation slot_memory_chroma;

    VkImage slot_image = nullptr;
    VkImageView slot_view = nullptr;
    uint32_t slot_index = 0;

    uint32_t output_query_index = 0;
    VkBuffer output_buffer = nullptr;
    VmaAllocation output_memory;

    VkDescriptorSet convert_descriptor = nullptr;
    lava::graphics_pipeline::ptr convert_pipeline;
    lava::render_pass::ptr convert_pass;

    VkDescriptorSet subsample_descriptor = nullptr;
    lava::graphics_pipeline::ptr subsample_pipeline;
    lava::render_pass::ptr subsample_pass;

    VkSemaphore render_semaphore = nullptr;
    VkSemaphore encode_semaphore = nullptr;
    VkCommandBuffer encode_command_buffer = nullptr;
    ExternFence::Ptr encode_fence;

    uint32_t frame_index = 0;
    EncoderFrameType type;
    EncoderCallback callback;
};

class Encoder
{
public:
    typedef std::shared_ptr<Encoder> Ptr;

public:
    Encoder() = default;

    bool create(lava::device_ptr device, const lava::renderer& renderer, asio::executor executor, const glm::uvec2& size, uint32_t frame_count);
    //NOTE: Assume all async operations has been completed.
    void destroy();

    void set_mode(EncoderMode mode);
    //NOTE: The number of frames between two key frames.
    void set_key_rate(uint32_t key_rate);
    //NOTE: The bit rate in Mbit per second.
    void set_bit_rate(double bit_rate);
    //NOTE: The frame rate in frames per second.
    void set_frame_rate(uint32_t frame_rate);
    //NOTE: The quality of i-frames in range from 0 to 51, where 0 means a lowest compression and 51 means highest compression.
    void set_quality_iframe(uint32_t quality);
    //NOTE: The quality of p-frames in range from 0 to 51, where 0 means a lowest compression and 51 means highest compression.
    void set_quality_pframe(uint32_t quality);

    bool encode_frame(VkCommandBuffer command_buffer, lava::renderer& renderer, lava::image::ptr image, VkImageLayout image_layout, EncoderCallback callback);

private:
    bool encode_control(lava::renderer& renderer, EncoderCallback callback);
    bool encode_config(lava::renderer& renderer, EncoderCallback callback);

    bool aquire_frame(EncoderFrame::Ptr& frame);
    void release_frame(EncoderFrame::Ptr frame);
    bool submit_frame(lava::device_ptr device, EncoderFrame::Ptr frame, lava::renderer& renderer);

    void write_image(lava::device_ptr device, EncoderFrame::Ptr frame, lava::image::ptr image);
    bool record_encoding(lava::device_ptr device, EncoderFrame::Ptr frame);
    void retrive_encoding(lava::device_ptr device, EncoderFrame::Ptr frame);

    bool create_profiles(lava::device_ptr device);
    bool compute_settings(const glm::uvec2& size);

    bool create_session(lava::device_ptr device, const glm::uvec2& size);
    bool create_parameters(lava::device_ptr device);
    bool bind_session_memory(lava::device_ptr device);

    bool create_sampler(lava::device_ptr device);
    bool create_pools(lava::device_ptr device, uint32_t frame_count);
    bool create_layouts(lava::device_ptr device);

    bool create_image_memory(lava::device_ptr device, VkImage image, VkImageAspectFlagBits image_aspect, VmaAllocation& memory);
    bool create_image_view(lava::device_ptr device, VkImage image, VkFormat image_format, VkImageView& image_view);

    bool create_input_images(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_slot_image(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_output_buffer(lava::device_ptr device, EncoderFrame::Ptr frame);

    bool create_convert_pass(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_convert_pipeline(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_subsample_pass(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_subsample_pipeline(lava::device_ptr device, EncoderFrame::Ptr frame);
    bool create_encode_resources(lava::device_ptr device, asio::executor executor, EncoderFrame::Ptr frame);

    void destroy_session(lava::device_ptr device);
    void destroy_resources(lava::device_ptr device);
    void destroy_frame(lava::device_ptr device, EncoderFrame::Ptr frame);

    bool check_encode_support(lava::device_ptr device, const lava::queue& queue) const;
    bool check_format_support(lava::device_ptr device, VkImageUsageFlags usage, VkFormat format) const;

    void convert_pass(VkCommandBuffer command_buffer);
    void subsample_pass(VkCommandBuffer command_buffer);

    void encode_pass_control_setup(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame);
    void encode_pass_control_command(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame);
    void encode_pass_config_setup(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame);
    void encode_pass_config_command(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame);
    void encode_pass_frame_setup(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame);
    void encode_pass_frame_command(VkCommandBuffer command_buffer, EncoderFrame::Ptr frame, const std::vector<EncoderFrame::Ptr>& reference_slots);

    void setup_slots();
    void process_slots(EncoderFrame::Ptr frame, std::vector<EncoderFrame::Ptr>& reference_slots);
    void invalidate_slots();

    bool is_key_frame() const;

private:
    uint32_t frame_count = 0;
    uint32_t frame_index = 0;

    bool config_change = true;
    bool control_change = true;

    EncoderMode setting_mode = ENCODER_MODE_CONSTANT_QUALITY;
    uint32_t setting_key_rate = 120;
    double setting_bit_rate = 2.0;          //NOTE: Bitrate in Mbit per seconds
    uint32_t setting_frame_rate = 90;
    uint32_t setting_quality_iframe = 51;
    uint32_t setting_quality_pframe = 51;
    
    uint32_t setting_reference_frames = 1;

    glm::uvec2 setting_block_count;
    glm::uvec2 setting_frame_crop;
    glm::uvec2 setting_image_size;
    uint32_t setting_buffer_size;

private:
    VkSampler sampler = nullptr;
    VkCommandPool command_pool = nullptr;
    VkQueryPool query_pool = nullptr;
    lava::descriptor::pool::ptr descriptor_pool;

    lava::descriptor::ptr descriptor_layout;
    lava::pipeline_layout::ptr pipeline_layout;

    VkSemaphore previous_semaphore = VK_NULL_HANDLE;

    std::mutex frame_mutex;
    std::vector<EncoderFrame::Ptr> frame_queue; //NOTE: Protected by frame_mutex
    std::vector<EncoderFrame::Ptr> frame_list;
    std::vector<EncoderFrame::Ptr> frame_slots;

private:
    lava::queue default_queue;
    std::optional<lava::queue> encode_queue;

    VkVideoSessionKHR video_session = nullptr;
    VkVideoSessionParametersKHR video_session_paremeters = nullptr;
    std::vector<VmaAllocation> video_session_memory;

    VkVideoProfileInfoKHR video_profile;
    VkVideoEncodeUsageInfoKHR usage_info;
    VkVideoEncodeH264ProfileInfoEXT encode_profile;

    VkVideoCapabilitiesKHR video_capabillities;
    VkVideoEncodeCapabilitiesKHR encode_capabillities;
    VkVideoEncodeH264CapabilitiesEXT codec_capabillities;

    const uint8_t sequence_parameter_id = 0;
    const uint8_t picture_parameter_id = 0;
};

bool setup_encoder(lava::device::create_param& parameters);

Encoder::Ptr make_encoder();
EncoderFrame::Ptr make_encoder_frame();