#include "latency.hpp"
#include <liblava/lava.hpp>
#include <fstream>
#include <chrono>
#include <filesystem>

LatencyStatistic::LatencyStatistic(const std::string& base_directory) : base_directory(base_directory)
{
    this->statistic_total_latency = make_statistic("Total Latency", 128);
    this->statistic_server_response = make_statistic("Server Response", 128);
    this->statistic_frame_decoded = make_statistic("Frame Decoded", 128);
    this->statistic_command_construct = make_statistic("Command Construct", 128);
    this->statistic_command_executed = make_statistic("Command Executed", 128);
}

void LatencyStatistic::interface()
{
    this->statistic_total_latency->interface();
    this->statistic_server_response->interface();
    this->statistic_frame_decoded->interface();
    this->statistic_command_construct->interface();
    this->statistic_command_executed->interface();
}

bool LatencyStatistic::write_to_file()
{
    if (!std::filesystem::exists(this->base_directory))
    {
        std::filesystem::create_directory(this->base_directory);
    }

    return this->write_captures();
}

void LatencyStatistic::submit_capture(const LatencyCapture& capture)
{
    double total_latency = capture.timestamp_command_executed - capture.timestamp_transform_query;
    double server_response = capture.timestamp_server_response - capture.timestamp_transform_query;
    double frame_decoded = capture.timestamp_frame_decoded - capture.timestamp_server_response;
    double command_construct = capture.timestamp_command_construct - capture.timestamp_frame_decoded;
    double command_executed = capture.timestamp_command_executed - capture.timestamp_command_construct;

    this->statistic_total_latency->add_sample(total_latency);
    this->statistic_server_response->add_sample(server_response);
    this->statistic_frame_decoded->add_sample(frame_decoded);
    this->statistic_command_construct->add_sample(command_construct);
    this->statistic_command_executed->add_sample(command_executed);

    this->captures.push_back(capture);
}

void LatencyStatistic::clear_captures()
{
    this->captures.clear();
}

bool LatencyStatistic::write_captures()
{
    std::string file_name = this->build_file_name();

    std::fstream file;
    file.open(this->base_directory + "/" + file_name, std::ios::out);

    if (!file.good())
    {
        lava::log()->error("can't write to file '" + file_name + "' for pass statistic");

        return false;
    }

    file << "frame_id; ";
    file << "frame_number; ";
    file << "transform_id; ";
    file << "timestamp_transform_query; ";
    file << "timestamp_server_response; ";
    file << "timestamp_frame_decoded; ";
    file << "timestamp_command_construct; ";
    file << "timestamp_command_executed; ";
    file << "total_latency" << std::endl;

    file << std::fixed << std::setprecision(3);

    for (const LatencyCapture& capture : this->captures)
    {
        double total_latency = capture.timestamp_command_executed - capture.timestamp_transform_query;

        file << capture.frame_id << "; ";
        file << capture.frame_number << "; ";
        file << capture.transform_id << "; ";
        file << capture.timestamp_transform_query << "; ";
        file << capture.timestamp_server_response << "; ";
        file << capture.timestamp_frame_decoded << "; ";
        file << capture.timestamp_command_construct << "; ";
        file << capture.timestamp_command_executed << "; ";
        file << total_latency << std::endl;
    }

    return true;
}

std::string LatencyStatistic::build_file_name()
{
    time_t system_time = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
    tm local_time = *localtime(&system_time);

    std::string name = "latency";
    name += "_" + std::to_string(local_time.tm_mday);
    name += "-" + std::to_string(local_time.tm_mon + 1);
    name += "-" + std::to_string(local_time.tm_year + 1900);
    name += "_" + std::to_string(local_time.tm_hour);
    name += "-" + std::to_string(local_time.tm_min);
    name += "-" + std::to_string(local_time.tm_sec);
    name += ".csv";

    return name;
}

LatencyStatistic::Ptr make_latency_statistic(const std::string& base_directory)
{
    return std::make_shared<LatencyStatistic>(base_directory);
}